##
# File: PdbxChemCompCategoryDefintions.py
# Date: 21-Feb-2012  John Westbrook
#
# Updates:
#
##
"""
A definitions of data categories used in the chemical component dictionary.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"


class PdbxChemCompCategoryDefinitions(object):
    _categoryInfo = [
        ("chem_comp", "key-value"),
        ("chem_comp_atom", "table"),
        ("chem_comp_bond", "table"),
        ("chem_comp_identifier", "table"),
        ("chem_comp_descriptor", "table"),
        ("pdbx_chem_comp_audit", "table"),
        ("pdbx_chem_comp_import", "table"),
        ("pdbx_chem_comp_atom_edit", "table"),
        ("pdbx_chem_comp_bond_edit", "table"),
    ]
    _cDict = {
        "chem_comp": [
            ("_chem_comp.id", "%s", "str", ""),
            ("_chem_comp.name", "%s", "str", ""),
            ("_chem_comp.type", "%s", "str", ""),
            ("_chem_comp.pdbx_type", "%s", "str", ""),
            ("_chem_comp.formula", "%s", "str", ""),
            ("_chem_comp.mon_nstd_parent_comp_id", "%s", "str", ""),
            ("_chem_comp.pdbx_synonyms", "%s", "str", ""),
            ("_chem_comp.pdbx_formal_charge", "%s", "str", ""),
            ("_chem_comp.pdbx_initial_date", "%s", "str", ""),
            ("_chem_comp.pdbx_modified_date", "%s", "str", ""),
            ("_chem_comp.pdbx_ambiguous_flag", "%s", "str", ""),
            ("_chem_comp.pdbx_release_status", "%s", "str", ""),
            ("_chem_comp.pdbx_replaced_by", "%s", "str", ""),
            ("_chem_comp.pdbx_replaces", "%s", "str", ""),
            ("_chem_comp.formula_weight", "%s", "str", ""),
            ("_chem_comp.one_letter_code", "%s", "str", ""),
            ("_chem_comp.three_letter_code", "%s", "str", ""),
            ("_chem_comp.pdbx_model_coordinates_details", "%s", "str", ""),
            ("_chem_comp.pdbx_model_coordinates_missing_flag", "%s", "str", ""),
            ("_chem_comp.pdbx_ideal_coordinates_details", "%s", "str", ""),
            ("_chem_comp.pdbx_ideal_coordinates_missing_flag", "%s", "str", ""),
            ("_chem_comp.pdbx_model_coordinates_db_code", "%s", "str", ""),
            ("_chem_comp.pdbx_subcomponent_list", "%s", "str", ""),
            ("_chem_comp.pdbx_processing_site", "%s", "str", ""),
        ],
        "chem_comp_atom": [
            ("_chem_comp_atom.comp_id", "%s", "str", ""),
            ("_chem_comp_atom.atom_id", "%s", "str", ""),
            ("_chem_comp_atom.alt_atom_id", "%s", "str", ""),
            ("_chem_comp_atom.type_symbol", "%s", "str", ""),
            ("_chem_comp_atom.charge", "%s", "str", ""),
            ("_chem_comp_atom.pdbx_align", "%s", "str", ""),
            ("_chem_comp_atom.pdbx_aromatic_flag", "%s", "str", ""),
            ("_chem_comp_atom.pdbx_leaving_atom_flag", "%s", "str", ""),
            ("_chem_comp_atom.pdbx_stereo_config", "%s", "str", ""),
            ("_chem_comp_atom.model_Cartn_x", "%s", "str", ""),
            ("_chem_comp_atom.model_Cartn_y", "%s", "str", ""),
            ("_chem_comp_atom.model_Cartn_z", "%s", "str", ""),
            ("_chem_comp_atom.pdbx_model_Cartn_x_ideal", "%s", "str", ""),
            ("_chem_comp_atom.pdbx_model_Cartn_y_ideal", "%s", "str", ""),
            ("_chem_comp_atom.pdbx_model_Cartn_z_ideal", "%s", "str", ""),
            ("_chem_comp_atom.pdbx_component_atom_id", "%s", "str", ""),
            ("_chem_comp_atom.pdbx_component_comp_id", "%s", "str", ""),
            ("_chem_comp_atom.pdbx_ordinal", "%s", "str", ""),
        ],
        "chem_comp_bond": [
            ("_chem_comp_bond.comp_id", "%s", "str", ""),
            ("_chem_comp_bond.atom_id_1", "%s", "str", ""),
            ("_chem_comp_bond.atom_id_2", "%s", "str", ""),
            ("_chem_comp_bond.value_order", "%s", "str", ""),
            ("_chem_comp_bond.pdbx_aromatic_flag", "%s", "str", ""),
            ("_chem_comp_bond.pdbx_stereo_config", "%s", "str", ""),
            ("_chem_comp_bond.pdbx_ordinal", "%s", "str", ""),
        ],
        "chem_comp_descriptor": [
            ("_pdbx_chem_comp_descriptor.comp_id", "%s", "str", ""),
            ("_pdbx_chem_comp_descriptor.type", "%s", "str", ""),
            ("_pdbx_chem_comp_descriptor.program", "%s", "str", ""),
            ("_pdbx_chem_comp_descriptor.program_version", "%s", "str", ""),
            ("_pdbx_chem_comp_descriptor.descriptor", "%s", "str", ""),
        ],
        "chem_comp_identifier": [
            ("_pdbx_chem_comp_identifier.comp_id", "%s", "str", ""),
            ("_pdbx_chem_comp_identifier.type", "%s", "str", ""),
            ("_pdbx_chem_comp_identifier.program", "%s", "str", ""),
            ("_pdbx_chem_comp_identifier.program_version", "%s", "str", ""),
            ("_pdbx_chem_comp_identifier.identifier", "%s", "str", ""),
        ],
        "pdbx_chem_comp_import": [
            ("_pdbx_chem_comp_identifier.comp_id", "%s", "str", ""),
            ("_pdbx_chem_comp_identifier.comp_alias_id", "%s", "str", ""),
            ("_pdbx_chem_comp_identifier.model_path", "%s", "str", ""),
        ],
        "pdbx_chem_comp_audit": [
            ("_pdbx_chem_comp_identifier.comp_id", "%s", "str", ""),
            ("_pdbx_chem_comp_identifier.action_type", "%s", "str", ""),
            ("_pdbx_chem_comp_identifier.date", "%s", "str", ""),
            ("_pdbx_chem_comp_identifier.processing_site", "%s", "str", ""),
            ("_pdbx_chem_comp_identifier.annotator", "%s", "str", ""),
            ("_pdbx_chem_comp_identifier.details", "%s", "str", ""),
        ],
        "pdbx_chem_comp_atom_edit": [
            ("_pdbx_chem_comp_atom_edit.ordinal", "%d", "int", ""),
            ("_pdbx_chem_comp_atom_edit.comp_id", "%s", "str", ""),
            ("_pdbx_chem_comp_atom_edit.atom_id", "%s", "str", ""),
            ("_pdbx_chem_comp_atom_edit.edit_op", "%s", "str", ""),
            ("_pdbx_chem_comp_atom_edit.edit_atom_id", "%s", "str", ""),
            ("_pdbx_chem_comp_atom_edit.edit_atom_value", "%s", "str", ""),
        ],
        "pdbx_chem_comp_bond_edit": [
            ("_pdbx_chem_comp_bond_edit.ordinal", "%d", "int", ""),
            ("_pdbx_chem_comp_bond_edit.comp_id_1", "%s", "str", ""),
            ("_pdbx_chem_comp_bond_edit.atom_id_1", "%s", "str", ""),
            ("_pdbx_chem_comp_bond_edit.comp_id_2", "%s", "str", ""),
            ("_pdbx_chem_comp_bond_edit.atom_id_2", "%s", "str", ""),
            ("_pdbx_chem_comp_bond_edit.edit_op", "%s", "str", ""),
            ("_pdbx_chem_comp_bond_edit.edit_bond_value", "%s", "str", ""),
        ],
    }



##
# File: PdbxChemCompDictUtil.py
# Date: 21-Feb-2012  John Westbrook
#
# Update:
#    1-Feb-2017 jdw unified with chem_ref_data -
##
"""
A collection of classes supporting maintenance methods on chemical
dictionaries and  persistent stores.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import sys
import os
import traceback

from mmcif_utils.persist.PdbxPersist import PdbxPersist
from mmcif_utils.persist.PdbxCoreIoAdapter import PdbxCoreIoAdapter as PdbxIoAdapter


class PdbxChemCompDictUtil(object):
    """Maintenance methods for creating and updating persistent stores of chemical dictionaries."""

    def __init__(self, verbose=True, log=sys.stderr):
        self.__verbose = verbose
        self.__debug = False
        self.__lfh = log

    def makeStoreFromFile(self, dictPath, storePath="chemcomp.db", minSize=10):
        """Create a new persistent store from input chemical dictionary file."""
        return self.__makeStoreFromFile(dictPath=dictPath, storePath=storePath, minSize=minSize)

    def makeStoreFromPathList(self, pathList, storePath="chemcomp.db"):
        """Create a new persistent store from a path list of chemical component definitions."""
        return self.__makeStoreFromPathList(pathList=pathList, storePath=storePath)

    def updateStoreByFile(self, pathList, storePath="chemcomp.db"):
        """Update the persistant store with the contents of the input path list"""
        return self.__updateStoreByFile(pathList, storePath=storePath)

    def updateStoreByObject(self, inpObject, containerName=None, containerType="data", storePath="chemcomp.db"):
        """Update the persistant store with the contents of the input object in the input named container."""
        return self.__updateStoreByObject(inpObject, containerName=containerName, containerType=containerType, storePath=storePath)

    def updateStoreByContainer(self, containerList, storePath="chemcomp.db"):
        """Update the persistant store with the contents of the input container list."""
        return self.__updateStoreByContainer(containerList=containerList, storePath=storePath)

    ##

    def __getFileSize(self, fPath):
        try:
            st = os.stat(fPath)
            return st.st_size
        except:  # noqa: E722 pylint: disable=bare-except
            return 0

    def __makeStoreFromFile(self, dictPath, storePath="chemcomp.db", minSize=10):
        """Internal method to create a new persistent store from input chemical dictionary file."""
        try:
            ok = False
            tmpPath = storePath + "-tmpstore"
            myReader = PdbxIoAdapter(self.__verbose, self.__lfh)
            ok = myReader.read(pdbxFilePath=dictPath)

            myPersist = PdbxPersist(self.__verbose, self.__lfh)
            myPersist.setContainerList(myReader.getContainerList())
            myPersist.store(dbFileName=tmpPath)
            if self.__getFileSize(tmpPath) > minSize:
                myPersist.moveStore(tmpPath, storePath)
                ok = True
                # if (self.__debug):
                #   indexD = myPersist.getIndex(dbFileName=storePath)
                #   self.__lfh.write("Persistent index dictionary %r\n" % indexD.items())
            else:
                ok = False
            return ok
        except:  # noqa: E722 pylint: disable=bare-except
            if self.__debug:
                traceback.print_exc(file=self.__lfh)
            return False

    def __makeStoreFromPathList(self, pathList, storePath="chemcomp.db", minSize=10):
        """Internal method to create a new persistent store from a path list
        of chemical component definitions.
        """
        try:
            ok = False
            # build the full container list from the input path list
            myReader = PdbxIoAdapter(self.__verbose, self.__lfh)
            for pth in pathList:
                ok = myReader.read(pdbxFilePath=pth)

            myPersist = PdbxPersist(self.__verbose, self.__lfh)
            myPersist.setContainerList(myReader.getContainerList())

            if self.__verbose:
                self.__lfh.write("Read completed for %d definitions\n" % len(pathList))

            tmpPath = storePath + "-tmpstore"
            myPersist.store(dbFileName=tmpPath)
            if self.__getFileSize(tmpPath) > minSize:
                ok = True
                myPersist.moveStore(tmpPath, storePath)
                #
                # if (self.__debug):
                #   indexD = myPersist.getIndex(dbFileName=storePath)
            else:
                ok = False
            return ok
        except:  # noqa: E722 pylint: disable=bare-except
            if self.__debug:
                traceback.print_exc(file=self.__lfh)
            return False

    def __updateStoreByFile(self, pathList, storePath="chemcomp.db"):
        """Internal method to update the persistant store with the contents of the input list of
        chemical component definition files.
        """
        try:
            myReader = PdbxIoAdapter(self.__verbose, self.__lfh)
            for rPath in pathList:
                _ok = myReader.read(pdbxFilePath=rPath)  # noqa: F841

                myPersist = PdbxPersist(self.__verbose, self.__lfh)
                myPersist.updateContainerList(dbFileName=storePath, containerList=myReader.getContainerList())
            #
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            if self.__debug:
                traceback.print_exc(file=self.__lfh)
            return False

    def __updateStoreByObject(self, inpObject, containerName=None, containerType="data", storePath="chemcomp.db"):
        """Internal method to update the persistant store with the contents of the input object in the input named container."""
        try:
            myPersist = PdbxPersist(self.__verbose, self.__lfh)
            myPersist.updateOneObject(inpObject, dbFileName=storePath, containerName=containerName, containerType=containerType)
        except:  # noqa: E722 pylint: disable=bare-except
            if self.__debug:
                traceback.print_exc(file=self.__lfh)
            return False

    def __updateStoreByContainer(self, containerList, storePath="chemcomp.db"):
        """Internal method to update the persistant store with the contents of the input container list."""
        try:
            myPersist = PdbxPersist(self.__verbose, self.__lfh)
            myPersist.updateContainerList(dbFileName=storePath, containerList=containerList)
        except:  # noqa: E722 pylint: disable=bare-except
            if self.__debug:
                traceback.print_exc(file=self.__lfh)
            return False



##
# File: PdbxChemCompPersist.py
# Date: 20-Feb-2012  John Westbrook
#
# Update:
#  21-Feb-2012 jdw adapted for chemcomputil repository
#  23-Feb-2012 jdw adapted for cc_dict_util repository
#   1-Feb-2017 jdw update imports to pdbx_v2
##
"""
A collection of access and iterator classes supporting chemical component dictionary data
extracted from persistent store.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import sys

# import traceback

from wwpdb.utils.cc_dict_util.persist.PdbxChemCompConstants import PdbxChemCompConstants


class PdbxCategoryItBase(object):
    """Base category iterator class."""

    def __init__(self, dataCategory, func, verbose=True, log=sys.stderr):  # pylint: disable=unused-argument
        self.__rL = dataCategory.getRowList()
        self.__func = func

    def get(self, index=0):
        try:
            return self.__rL[index]
        except:  # noqa: E722 pylint: disable=bare-except
            return []

    def __iter__(self):
        return self.forward()

    def forward(self):
        # Forward generator
        current_row = 0
        while current_row < len(self.__rL):
            row = self.__rL[current_row]
            current_row += 1
            yield self.__func(row)

    def reverse(self):
        # The reverse generator
        current_row = len(self.__rL)
        while current_row > 0:
            current_row -= 1
            yield self.__func(self.__rL[current_row])


class PdbxChemCompIt(PdbxCategoryItBase):
    def __init__(self, dataCategory, verbose=True, log=sys.stderr):
        o = PdbxChemCompPersist([], attributeNameList=dataCategory.getAttributeList(), verbose=verbose, log=log)
        super(PdbxChemCompIt, self).__init__(dataCategory, o.set, verbose, log)


class PdbxChemCompAtomIt(PdbxCategoryItBase):
    def __init__(self, dataCategory, verbose=True, log=sys.stderr):
        o = PdbxChemCompAtomPersist([], attributeNameList=dataCategory.getAttributeList(), verbose=verbose, log=log)
        super(PdbxChemCompAtomIt, self).__init__(dataCategory, o.set, verbose, log)


class PdbxChemCompBondIt(PdbxCategoryItBase):
    def __init__(self, dataCategory, verbose=True, log=sys.stderr):
        o = PdbxChemCompBondPersist([], attributeNameList=dataCategory.getAttributeList(), verbose=verbose, log=log)
        super(PdbxChemCompBondIt, self).__init__(dataCategory, o.set, verbose, log)


class PdbxChemCompDescriptorIt(PdbxCategoryItBase):
    def __init__(self, dataCategory, verbose=True, log=sys.stderr):
        o = PdbxChemCompDescriptorPersist([], attributeNameList=dataCategory.getAttributeList(), verbose=verbose, log=log)
        super(PdbxChemCompDescriptorIt, self).__init__(dataCategory, o.set, verbose, log)


class PdbxChemCompIdentifierIt(PdbxCategoryItBase):
    def __init__(self, dataCategory, verbose=True, log=sys.stderr):
        o = PdbxChemCompIdentifierPersist([], attributeNameList=dataCategory.getAttributeList(), verbose=verbose, log=log)
        super(PdbxChemCompIdentifierIt, self).__init__(dataCategory, o.set, verbose, log)


class PdbxChemCompAuditIt(PdbxCategoryItBase):
    def __init__(self, dataCategory, verbose=True, log=sys.stderr):
        o = PdbxChemCompAuditPersist([], attributeNameList=dataCategory.getAttributeList(), verbose=verbose, log=log)
        super(PdbxChemCompAuditIt, self).__init__(dataCategory, o.set, verbose, log)


class PdbxChemCompPersist(object):
    """Accessor methods chemical component attributes."""

    def __init__(self, rowData, attributeNameList, verbose=True, log=sys.stderr):  # pylint: disable=unused-argument
        self.__rowData = rowData
        self.__attributeNameList = attributeNameList

    def set(self, rowData=None):
        self.__rowData = rowData
        return self

    def __getAttribute(self, name):
        try:
            i = self.__attributeNameList.index(name)
            return self.__rowData[i]
        except:  # noqa: E722 pylint: disable=bare-except
            return None

    def getId(self):
        return self.__getAttribute("id")

    def getName(self):
        return self.__getAttribute("name")

    def getType(self):
        return self.__getAttribute("type")

    def getPdbxType(self):
        return self.__getAttribute("pdbx_type")

    def getFormula(self):
        return self.__getAttribute("formula")

    def getSynonyms(self):
        return self.__getAttribute("pdbx_synonyms")

    def getFormalCharge(self):
        return self.__getAttribute("pdbx_formal_charge")

    def getModificationDate(self):
        return self.__getAttribute("pdbx_modified_date")

    def getInitialDate(self):
        return self.__getAttribute("pdbx_initial_date")

    def getReleaseStatus(self):
        return self.__getAttribute("pdbx_release_status")

    def getFormulaWeight(self):
        return self.__getAttribute("formula_weight")

    def getSubComponentList(self):
        return self.__getAttribute("pdbx_subcomponent_list")

    def getAmbiguousFlag(self):
        return self.__getAttribute("pdbx_ambiguous_flag")

    def getProcessingSite(self):
        return self.__getAttribute("pdbx_processing_site")

    def getReplacesId(self):
        return self.__getAttribute("pdbx_replaces")

    def getReplacesById(self):
        return self.__getAttribute("pdbx_replaced_by")

    def getNstdParentId(self):
        return self.__getAttribute("mon_nstd_parent_comp_id")

    def getOneLetterCode(self):
        return self.__getAttribute("one_letter_code")

    def getThreeLetterCode(self):
        return self.__getAttribute("three_letter_code")

    def getModelCoordinatesPdbCode(self):
        return self.__getAttribute("pdbx_model_coordinates_db_code")

    def getMissingModelCoordinates(self):
        return self.__getAttribute("pdbx_model_coordinates_missing_flag")

    def getMissingIdealCoordinates(self):
        return self.__getAttribute("pdbx_ideal_coordinates_missing_flag")


class PdbxChemCompAtomPersist(PdbxChemCompConstants):
    """Accessor methods chemical component atom attributes."""

    def __init__(self, rowData, attributeNameList, verbose=True, log=sys.stderr):  # pylint: disable=unused-argument
        super(PdbxChemCompAtomPersist, self).__init__()
        self.__rowData = rowData
        self.__attributeNameList = attributeNameList

    def __getAttribute(self, name):
        try:
            i = self.__attributeNameList.index(name)
            return self.__rowData[i]
        except:  # noqa: E722 pylint: disable=bare-except
            return None

    def set(self, rowData=None):
        self.__rowData = rowData
        return self

    def getName(self):
        return self.__getAttribute("atom_id")

    def isChiral(self):
        return self.__getAttribute("pdbx_stereo_config") != "N"

    def getType(self):
        return self.__getAttribute("type_symbol")

    def getLeavingAtomFlag(self):
        return self.__getAttribute("pdbx_leaving_atom_flag")

    def getAtNo(self):
        try:
            tyU = str(self.getType()).upper()
            if (tyU == "D") or (tyU == "T"):
                tyU = "H"
            return self._periodicTable.index(tyU) + 1
        except:  # noqa: E722 pylint: disable=bare-except
            # traceback.print_exc(file=self.__lfh)
            return 0

    def getIsotope(self):
        ty = self.getType()
        if ty == "D":
            return 2
        elif ty == "T":
            return 3
        else:
            return 0

    def isAromatic(self):
        return self.__getAttribute("pdbx_aromatic_flag") != "N"

    def getCIPStereo(self):
        return self.__getAttribute("pdbx_stereo_config")

    def getFormalCharge(self):
        try:
            return int(self.__getAttribute("charge"))
        except:  # noqa: E722 pylint: disable=bare-except
            return 0

    def hasModelCoordinates(self):
        x, y, z = self.getModelCoordinates()
        # x=self.__getAttribute('model_Cartn_x')
        # y=self.__getAttribute('model_Cartn_y')
        # z=self.__getAttribute('model_Cartn_z')
        #
        return (x is not None) and (y is not None) and (z is not None)

    def hasIdealCoordinates(self):
        x, y, z = self.getIdealCoordinates()
        # x=self.__getAttribute('pdbx_model_Cartn_x_ideal')
        # y=self.__getAttribute('pdbx_model_Cartn_y_ideal')
        # z=self.__getAttribute('pdbx_model_Cartn_z_ideal')
        #
        return (x is not None) and (y is not None) and (z is not None)

    def getModelCoordinates(self):
        """Returns (x,y,z)"""
        try:
            x = float(self.__getAttribute("model_Cartn_x"))
            y = float(self.__getAttribute("model_Cartn_y"))
            z = float(self.__getAttribute("model_Cartn_z"))
            return (x, y, z)
        except:  # noqa: E722 pylint: disable=bare-except
            return (None, None, None)

    def getIdealCoordinates(self):
        """Returns (x,y,z)"""
        try:
            x = float(self.__getAttribute("pdbx_model_Cartn_x_ideal"))
            y = float(self.__getAttribute("pdbx_model_Cartn_y_ideal"))
            z = float(self.__getAttribute("pdbx_model_Cartn_z_ideal"))
            return (x, y, z)
        except:  # noqa: E722 pylint: disable=bare-except
            return (None, None, None)

    def dump(self, ofh):
        ofh.write("PdbxChemCompAtomPersist(dump) %r\n" % self.__rowData)


class PdbxChemCompBondPersist(object):
    """Accessor methods chemical component bond attributes."""

    def __init__(self, rowData, attributeNameList, verbose=True, log=sys.stderr):  # pylint: disable=unused-argument
        self.__rowData = rowData
        self.__attributeNameList = attributeNameList

    def __getAttribute(self, name):
        try:
            i = self.__attributeNameList.index(name)
            return self.__rowData[i]
        except:  # noqa: E722 pylint: disable=bare-except
            return None

    def set(self, rowData=None):
        self.__rowData = rowData
        return self

    def getBond(self):
        """Returns (atomI,atomJ) atom ids from the atom list."""
        return (self.__getAttribute("atom_id_1"), self.__getAttribute("atom_id_2"))

    def getType(self):
        return self.__getAttribute("value_order")

    def getIntegerType(self):
        bT = self.__getAttribute("value_order")
        if bT == "SING":
            return 1
        elif bT == "DOUB":
            return 2
        elif bT == "TRIP":
            return 3
        elif bT == "QUAD":
            return 4
        else:
            return 0

    def isAromatic(self):
        return self.__getAttribute("pdbx_aromatic_flag") == "Y"

    def getStereo(self):
        return self.__getAttribute("pdbx_stereo_config")

    def hasStereo(self):
        return self.__getAttribute("pdbx_stereo_config") != "N"

    def dump(self, ofh):
        ofh.write("PdbxChemCompBondPersist(dump) %r\n" % self.__rowData)


class PdbxChemCompIdentifierPersist(object):
    """Accessor methods chemical component identifier attributes."""

    def __init__(self, rowData, attributeNameList, verbose=True, log=sys.stderr):  # pylint: disable=unused-argument
        self.__rowData = rowData
        self.__attributeNameList = attributeNameList

    def __getAttribute(self, name):
        try:
            i = self.__attributeNameList.index(name)
            return self.__rowData[i]
        except:  # noqa: E722 pylint: disable=bare-except
            return None

    def set(self, rowData=None):
        self.__rowData = rowData
        return self

    def getIdentifier(self):
        """Returns the value of the identifier."""
        return self.__getAttribute("identifier")

    def getType(self):
        return self.__getAttribute("type")

    def getProgram(self):
        return self.__getAttribute("program")

    def getProgramVersion(self):
        return self.__getAttribute("program_version")

    def dump(self, ofh):
        ofh.write("PdbxChemCompIdentifierPersist(dump) %r\n" % self.__rowData)


class PdbxChemCompDescriptorPersist(object):
    """Accessor methods chemical component descriptor  attributes."""

    def __init__(self, rowData, attributeNameList, verbose=True, log=sys.stderr):  # pylint: disable=unused-argument
        self.__rowData = rowData
        self.__attributeNameList = attributeNameList

    def __getAttribute(self, name):
        try:
            i = self.__attributeNameList.index(name)
            return self.__rowData[i]
        except:  # noqa: E722 pylint: disable=bare-except
            return None

    def set(self, rowData=None):
        self.__rowData = rowData
        return self

    def getDescriptor(self):
        """Returns the value of the descriptor."""
        return self.__getAttribute("descriptor")

    def getType(self):
        return self.__getAttribute("type")

    def getProgram(self):
        return self.__getAttribute("program")

    def getProgramVersion(self):
        return self.__getAttribute("program_version")

    def dump(self, ofh):
        ofh.write("PdbxChemCompDescriptorPersist(dump) %r\n" % self.__rowData)


class PdbxChemCompAuditPersist(object):
    """Accessor methods chemical component audit details."""

    def __init__(self, rowData, attributeNameList, verbose=True, log=sys.stderr):  # pylint: disable=unused-argument
        self.__rowData = rowData
        self.__attributeNameList = attributeNameList

    def __getAttribute(self, name):
        try:
            i = self.__attributeNameList.index(name)
            return self.__rowData[i]
        except:  # noqa: E722 pylint: disable=bare-except
            return None

    def set(self, rowData=None):
        self.__rowData = rowData
        return self

    def getActionType(self):
        """Returns the value of the action type."""
        return self.__getAttribute("action_type")

    def getDate(self):
        """Returns the value of audit date."""
        return self.__getAttribute("date")

    def getProcessingSite(self):
        """Returns the value of processing site."""
        return self.__getAttribute("processing_site")

    def getAnnotator(self):
        """Returns the value of audit annotator."""
        return self.__getAttribute("annotator")

    def getDetails(self):
        """Returns the value of audit details."""
        return self.__getAttribute("details")



##
#  File:  TimeoutMultiProc.py
#  Date:  17-Nov-2011  Jdw
#
##

"""
Code to timeout with processes.

>>> @timeout(.5)
... def sleep(x):
...     print "ABOUT TO SLEEP {0} SECONDS".format(x)
...     time.sleep(x)
...     return x

>>> sleep(1)
Traceback (most recent call last):
   ...
TimeoutException: timed out after 0 seconds

>>> sleep(.2)
0.2

>>> @timeout(.5)
... def exc():
...     raise Exception('Houston we have problems!')

>>> exc()
Traceback (most recent call last):
   ...
Exception: Houston we have problems!

"""
import multiprocessing
import time

# import logging
import signal
from functools import wraps


class TimeoutException(Exception):
    def __init__(self, value):
        super(TimeoutException, self).__init__(value)

        self.value = value

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return repr(self.value)


class RunableProcessing(multiprocessing.Process):  # pragma: no cover
    def __init__(self, func, *args, **kwargs):
        self.queue = multiprocessing.Queue(maxsize=1)
        args = (func,) + args
        multiprocessing.Process.__init__(self, target=self.run_func, args=args, kwargs=kwargs)
        # logger = multiprocessing.log_to_stderr()
        # logger.setLevel(logging.INFO)

    def run_func(self, func, *args, **kwargs):
        try:
            result = func(*args, **kwargs)
            self.queue.put((True, result))
        except Exception as e:
            self.queue.put((False, e))

    def done(self):
        return self.queue.full()

    def result(self):
        return self.queue.get()


def timeout(seconds, message="Function call timed out"):
    def wrapper(function):
        def _handleTimeout(signum, frame):
            raise TimeoutException(message)

        @wraps(function)
        def wrapped(*args, **kwargs):
            signal.signal(signal.SIGALRM, _handleTimeout)
            signal.alarm(seconds)
            try:
                result = function(*args, **kwargs)
            finally:
                signal.alarm(0)
            return result

        return wrapped

    return wrapper


def timeoutMp(seconds, force_kill=True):  # pragma: no cover
    def wrapper(function):
        @wraps(function)
        def inner(*args, **kwargs):
            now = time.time()
            proc = RunableProcessing(function, *args, **kwargs)
            proc.start()
            proc.join(seconds)
            if proc.is_alive():
                if force_kill:
                    proc.terminate()
                runtime = int(time.time() - now)
                raise TimeoutException("timed out after {0} seconds".format(runtime))
            assert proc.done()
            success, result = proc.result()
            if success:
                return result
            else:
                raise result

        return inner

    return wrapper


if __name__ == "__main__":  # pragma: no cover
    import doctest

    doctest.testmod()



##
# File: PdbxChemCompDictIndex.py
# Date: 21-Feb-2012  John Westbrook
#
# Update:
#
#  21-Apr-2013  jdw  Add indices of parent/child relationships for modified residues
#  17-Mar-2016  jdw  Update and move to chem_ref_data module -
#   1-Feb-2017  jdw  Move search functions to ChemCompIndexSearchUtils.py
#   1-Feb-2017  jdw  Unified with chem_ref_data -
#   1-Feb-2017  jdw  Expand index content
#   1-May-2017  jdw  Index all SMILES -
#
##
"""
A collection of classes for building simplified indices of the
serialized content in the chemical component dictionary.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import sys
import traceback

try:
    import cPickle as pickle
except ImportError:
    import pickle

from mmcif_utils.persist.PdbxPersist import PdbxPersist
from wwpdb.utils.cc_dict_util.persist.PdbxChemCompPersist import PdbxChemCompIt, PdbxChemCompAtomIt, PdbxChemCompDescriptorIt, PdbxChemCompIdentifierIt


class PdbxChemCompDictIndex(object):
    """Builds simplified indices of the serialized content in the chemical component dictionary."""

    def __init__(self, verbose=True, log=sys.stderr):
        self.__verbose = verbose
        self.__debug = False
        self.__lfh = log

    def makeIndex(self, storePath="chemcomp.db", indexPath="chemcomp-index.pic"):
        """Create a search index from the contents of a persistent store of
        a chemical dictionary.  Store the index in indexPath.
        """
        return self.__makeIndex(storePath=storePath, indexPath=indexPath)

    def makeParentComponentIndex(self, storePath="chemcomp.db", indexPath="chemcomp-parent-index.pic"):
        """Create a search index for parent components from the contents of a persistent store of
        a chemical dictionary.  Store the index in indexPath.
        """
        return self.__makeParentIndex(storePath=storePath, indexPath=indexPath)

    def readIndex(self, indexPath="chemcomp-index.pic"):
        """Read and return search index."""
        return self.__readIndex(indexPath=indexPath)

    def readParentComponentIndex(self, indexPath="chemcomp-parent-index.pic"):
        """Read and return the parent component search index."""
        return self.__readParentIndex(indexPath=indexPath)

    def __makeIndex(self, storePath, indexPath):
        """Read serialized component dictionary and build a search index.

        Index is a dictionary of selected items such as name, synonyns, formula, status and
        descriptors.  Index is stored as a pickled dictionary.
        """
        ccIdx = {}
        try:
            myPersist = PdbxPersist(self.__verbose, self.__lfh)
            myPersist.open(dbFileName=storePath)
            containerList = myPersist.getStoreContainerIndex()
            #
            for ccId in containerList:
                #
                d = {}
                d["ccId"] = ccId
                d["nameList"] = []
                d["typeCounts"] = {}
                d["InChI"] = None
                d["InChIKey"] = None
                d["InChIKey14"] = None
                d["smiles"] = None
                d["smilesList"] = []
                d["smilesStereo"] = None
                d["releaseStatus"] = None
                d["subcomponentList"] = None
                #
                d["name"] = None
                d["type"] = None
                d["formula"] = None
                d["formulaWeight"] = None
                #
                nameList = []
                dC = myPersist.fetchObject(containerName=ccId, objectName="chem_comp")
                if dC is not None:
                    rowIt = PdbxChemCompIt(dC, self.__verbose, self.__lfh)
                    for row in rowIt:
                        name = row.getName()
                        synonyms = row.getSynonyms()
                        d["releaseStatus"] = row.getReleaseStatus()
                        d["subcomponentList"] = row.getSubComponentList()
                        d["name"] = name
                        d["synonyms"] = synonyms
                        d["type"] = row.getType()
                        d["formula"] = row.getFormula()
                        d["formulaWeight"] = row.getFormulaWeight()
                        d["ambiguousFlag"] = row.getAmbiguousFlag()

                    nameList.append(name)
                    if synonyms is not None:
                        if ";" in synonyms:
                            sList = synonyms.split(";")
                            nameList.extend(sList)
                        else:
                            nameList.append(synonyms)

                # Compute element/type counts directly from the definition atom list
                typeCounts = {}
                dC = myPersist.fetchObject(containerName=ccId, objectName="chem_comp_atom")
                if dC is not None:
                    rowIt = PdbxChemCompAtomIt(dC, self.__verbose, self.__lfh)
                    for row in rowIt:
                        aType = row.getType()
                        if row.getType() not in typeCounts:
                            typeCounts[aType] = 1
                        else:
                            typeCounts[aType] += 1
                    d["typeCounts"] = typeCounts

                #
                dC = myPersist.fetchObject(containerName=ccId, objectName="pdbx_chem_comp_descriptor")
                if dC is not None:
                    rowIt = PdbxChemCompDescriptorIt(dC, self.__verbose, self.__lfh)
                    for row in rowIt:
                        des = row.getDescriptor()
                        desType = row.getType()
                        desProgram = row.getProgram()
                        if desType.startswith("SMILES"):
                            d["smilesList"].append(des)
                        if "OpenEye" in desProgram:
                            if desType == "SMILES_CANNONICAL":
                                d["smilesStereo"] = des
                            elif desType == "SMILES":
                                d["smiles"] = des
                        elif "InChI" in desProgram:
                            if desType == "InChI":
                                d["InChI"] = des
                            elif desType == "InChIKey":
                                d["InChIKey"] = des
                                d["InChIKey14"] = des[:14]
                #

                dC = myPersist.fetchObject(containerName=ccId, objectName="pdbx_chem_comp_identifier")
                if dC is not None:
                    rowIt = PdbxChemCompIdentifierIt(dC, self.__verbose, self.__lfh)
                    for row in rowIt:
                        iden = row.getIdentifier()
                        idenType = row.getType()
                        # idenProgram = row.getProgram()
                        if "SYSTEMATIC" in idenType:
                            nameList.append(iden)

                d["nameList"] = nameList
                ccIdx[ccId] = d

            myPersist.close()
            with open(indexPath, "wb") as fout:
                # Maintain backwards compatibility with python 2
                pickle.dump(ccIdx, fout, 2)

        except:  # noqa: E722 pylint: disable=bare-except
            if self.__verbose:
                self.__lfh.write("PdbxChemCompDictIndex(__makeIndex) index creation failed for %s index %s\n" % (storePath, indexPath))
            if self.__debug:
                traceback.print_exc(file=self.__lfh)

        return ccIdx

    def __makeParentIndex(self, storePath, indexPath):
        """Read serialized component dictionary and indices of parent/child relationships for modified residues.

        Index is stored as a pickled dictionary in indexPath.
        """
        pD = {}
        cD = {}
        try:
            myPersist = PdbxPersist(self.__verbose, self.__lfh)
            myPersist.open(dbFileName=storePath)
            containerList = myPersist.getStoreContainerIndex()
            #

            for ccId in containerList:
                #
                dC = myPersist.fetchObject(containerName=ccId, objectName="chem_comp")
                if dC is not None:
                    rowIt = PdbxChemCompIt(dC, self.__verbose, self.__lfh)
                    for row in rowIt:
                        pCompId = row.getNstdParentId()
                        compId = row.getId()
                        #
                        if (pCompId is not None) and (len(pCompId) > 0) and (pCompId not in ["?", "."]):

                            if "," in pCompId:
                                pList = pCompId.split(",")
                                cD[compId] = pList
                            else:
                                if len(pCompId) > 3:
                                    self.__lfh.write("PdbxChemCompDictIndex(__makeParentIndex) compId %s parent compId %s\n" % (compId, pCompId))
                                else:
                                    pList = [pCompId]
                                    if pCompId not in pD:
                                        pD[pCompId] = []
                                    pD[pCompId].append(compId)
                                    #
                                    cD[compId] = pList

            myPersist.close()
            #
            ofh = open(indexPath, "wb")
            # Compatbility with python 2 and not pickle.HIGHEST_PROTOCOL
            pickle.dump(pD, ofh, 2)
            pickle.dump(cD, ofh, 2)
            ofh.close()

        except:  # noqa: E722 pylint: disable=bare-except
            if self.__verbose:
                self.__lfh.write("PdbxChemCompDictIndex(__makeParentIndex) parent index creation failed for %s index %s\n" % (storePath, indexPath))
            if self.__debug:
                traceback.print_exc(file=self.__lfh)

        return pD, cD

    def __readParentIndex(self, indexPath):
        """Internal method to recover the pickled index file."""
        pD = {}
        cD = {}
        try:
            ifh = open(indexPath, "rb")
            pD = pickle.load(ifh)
            cD = pickle.load(ifh)
            ifh.close()
            return pD, cD
        except:  # noqa: E722 pylint: disable=bare-except
            pass
        return pD, cD

    def __readIndex(self, indexPath):
        """Internal method to recover the pickled index file."""
        try:
            with open(indexPath, "rb") as fin:
                return pickle.load(fin)
        except:  # noqa: E722 pylint: disable=bare-except
            return {}
